@@ -95,6 +95,15 @@ class S3FileSystem(object):
     use_ssl : bool (True)
         Whether to use SSL in connections to S3; may be faster without, but
         insecure
+    client_kwargs : dict of paramaters for the boto3 client
+    requester_pays : bool (False)
+        If RequesterPays buckets are supported.
+    default_block_size: None, int
+        If given, the default block size value used for ``open()``, if no
+        specific value is given at all time. The built-in default is 5MB.
+    default_fill_cache: Bool (True)
+        Whether to use cache filling with open by default. Refer to
+        ``S3File.open``.
     kwargs : other parameters for boto3 session
 
     Examples
@@ -111,15 +120,26 @@ class S3FileSystem(object):
     _singleton = [None]
     connect_timeout = 5
     read_timeout = 15
+    default_block_size = 5 * 2**20
 
     def __init__(self, anon=False, key=None, secret=None, token=None,
-                 use_ssl=True, **kwargs):
+                 use_ssl=True, client_kwargs=None, requester_pays=False,
+                 default_block_size=None, default_fill_cache=True,
+                 **kwargs):
         self.anon = anon
         self.key = key
         self.secret = secret
         self.token = token
         self.kwargs = kwargs
+
+        if client_kwargs is None:
+            client_kwargs = {}
+        if default_block_size is not None:
+            self.default_block_size = default_block_size
+        self.default_fill_cache = default_fill_cache
+        self.client_kwargs = client_kwargs
         self.dirs = {}
+        self.req_kw = {'RequestPayer': 'requester'} if requester_pays else {}
         self.use_ssl = use_ssl
         self.s3 = self.connect()
         S3FileSystem._singleton[0] = self
@@ -144,12 +164,14 @@ class S3FileSystem(object):
         refresh : bool (True)
             Whether to use cached filelists, if already read
         """
-        anon, key, secret, kwargs, token, ssl = (self.anon, self.key,
-                            self.secret, self.kwargs, self.token, self.use_ssl)
+        anon, key, secret, kwargs, ckwargs, token, ssl = \
+              (self.anon, self.key, self.secret, self.kwargs,
+               self.client_kwargs, self.token, self.use_ssl)
 
         # Include the current PID in the connection key so that different
         # SSL connections are made for each process.
-        tok = tokenize(anon, key, secret, kwargs, token, ssl, os.getpid())
+        tok = tokenize(anon, key, secret, kwargs, ckwargs, token,
+                       ssl, os.getpid())
         if refresh:
             self._conn.pop(tok, None)
         logger.debug("Open S3 connection.  Anonymous: %s", self.anon)
@@ -166,7 +188,8 @@ class S3FileSystem(object):
             self.session = boto3.Session(self.key, self.secret, self.token,
                                          **self.kwargs)
         if tok not in self._conn:
-            s3 = self.session.client('s3', config=conf, use_ssl=ssl)
+            s3 = self.session.client('s3', config=conf, use_ssl=ssl,
+                                     **self.client_kwargs)
             self._conn[tok] = s3
         return self._conn[tok]
 
@@ -206,7 +229,8 @@ class S3FileSystem(object):
         self.__dict__.update(state)
         self.s3 = self.connect()
 
-    def open(self, path, mode='rb', block_size=5 * 1024 ** 2, acl=''):
+    def open(self, path, mode='rb', block_size=None, acl='',
+             fill_cache=None):
         """ Open a file for reading or writing
 
         Parameters
@@ -217,11 +241,21 @@ class S3FileSystem(object):
             One of 'rb' or 'wb'
         block_size: int
             Size of data-node blocks if reading
+        fill_cache: bool
+            If seeking to new a part of the file beyond the current buffer,
+            with this True, the buffer will be filled between the sections to
+            best support random access. When reading only a few specific chunks
+            out of a file, performance may be better if False.
         """
+        if block_size is None:
+            block_size = self.default_block_size
+        if fill_cache is None:
+            fill_cache = self.default_fill_cache
         if 'b' not in mode:
             raise NotImplementedError("Text mode not supported, use mode='%s'"
                                       " and manage bytes" % (mode[0] + 'b'))
-        return S3File(self, path, mode, block_size=block_size, acl=acl)
+        return S3File(self, path, mode, block_size=block_size, acl=acl,
+                      fill_cache=fill_cache)
 
     def _lsdir(self, path, refresh=False):
         if path.startswith('s3://'):
@@ -232,7 +266,8 @@ class S3FileSystem(object):
         if path not in self.dirs or refresh:
             try:
                 pag = self.s3.get_paginator('list_objects')
-                it = pag.paginate(Bucket=bucket, Prefix=prefix, Delimiter='/')
+                it = pag.paginate(Bucket=bucket, Prefix=prefix, Delimiter='/',
+                                  **self.req_kw)
                 files = []
                 dirs = None
                 for i in it:
@@ -320,7 +355,8 @@ class S3FileSystem(object):
         else:
             try:
                 bucket, key = split_path(path)
-                out = self.s3.head_object(Bucket=bucket, Key=key)
+                out = self.s3.head_object(Bucket=bucket, Key=key,
+                                          **self.req_kw)
                 out = {'ETag': out['ETag'], 'Key': '/'.join([bucket, key]),
                        'LastModified': out['LastModified'],
                        'Size': out['ContentLength'], 'StorageClass': "STANDARD"}
@@ -451,7 +487,6 @@ class S3FileSystem(object):
                                        Params={'Bucket': bucket, 'Key': key},
                                        ExpiresIn=expires)
 
-
     def get(self, path, filename):
         """ Stream data from file at path to local filename """
         with self.open(path, 'rb') as f:
@@ -677,6 +712,11 @@ class S3File(object):
         S3 key to access
     blocksize : int
         read-ahead size for finding delimiters
+    fill_cache: bool
+        If seeking to new a part of the file beyond the current buffer,
+        with this True, the buffer will be filled between the sections to
+        best support random access. When reading only a few specific chunks
+        out of a file, performance may be better if False.
 
     Examples
     --------
@@ -689,7 +729,8 @@ class S3File(object):
     S3FileSystem.open: used to create ``S3File`` objects
     """
 
-    def __init__(self, s3, path, mode='rb', block_size=5 * 2 ** 20, acl=""):
+    def __init__(self, s3, path, mode='rb', block_size=5 * 2 ** 20, acl="",
+                 fill_cache=True):
         self.mode = mode
         if mode not in {'rb', 'wb', 'ab'}:
             raise NotImplementedError("File mode must be {'rb', 'wb', 'ab'}, not %s" % mode)
@@ -709,6 +750,7 @@ class S3File(object):
         self.trim = True
         self.mpu = None
         self.acl = acl
+        self.fill_cache = fill_cache
         if mode in {'wb', 'ab'}:
             if acl and acl not in key_acls:
                 raise ValueError('ACL not in %s', key_acls)
@@ -819,17 +861,24 @@ class S3File(object):
             self.start = start
             self.end = end + self.blocksize
             self.cache = _fetch_range(self.s3.s3, self.bucket, self.key,
-                                      start, self.end)
+                                      start, self.end, req_kw=self.s3.req_kw)
         if start < self.start:
+            if not self.fill_cache and end + self.blocksize < self.start:
+                self.start, self.end = None, None
+                return self._fetch(start, end)
             new = _fetch_range(self.s3.s3, self.bucket, self.key,
-                               start, self.start)
+                               start, self.start, req_kw=self.s3.req_kw)
             self.start = start
             self.cache = new + self.cache
         if end > self.end:
             if self.end > self.size:
                 return
+            if not self.fill_cache and start > self.end:
+                self.start, self.end = None, None
+                return self._fetch(start, end)
             new = _fetch_range(self.s3.s3, self.bucket, self.key,
-                               self.end, end + self.blocksize)
+                               self.end, end + self.blocksize,
+                               req_kw=self.s3.req_kw)
             self.end = end + self.blocksize
             self.cache = self.cache + new
 
@@ -863,7 +912,7 @@ class S3File(object):
         """
         Write data to buffer.
 
-        Buffer only sent to S3 on flush() or if buffer is bigger than blocksize.
+        Buffer only sent to S3 on flush() or if buffer is greater than or equal to blocksize.
 
         Parameters
         ----------
@@ -876,7 +925,7 @@ class S3File(object):
             raise ValueError('I/O operation on closed file.')
         out = self.buffer.write(ensure_writable(data))
         self.loc += out
-        if self.buffer.tell() > self.blocksize:
+        if self.buffer.tell() >= self.blocksize:
             self.flush()
         return out
 
@@ -993,12 +1042,16 @@ class S3File(object):
         self.close()
 
 
-def _fetch_range(client, bucket, key, start, end, max_attempts=10):
+def _fetch_range(client, bucket, key, start, end, max_attempts=10,
+                 req_kw=None):
+    if req_kw is None:
+        req_kw = {}
     logger.debug("Fetch: %s/%s, %s-%s", bucket, key, start, end)
     for i in range(max_attempts):
         try:
             resp = client.get_object(Bucket=bucket, Key=key,
-                                     Range='bytes=%i-%i' % (start, end - 1))
+                                     Range='bytes=%i-%i' % (start, end - 1),
+                                     **req_kw)
             return resp['Body'].read()
         except S3_RETRYABLE_ERRORS as e:
             logger.debug('Exception %e on S3 download, retrying', e,
@@ -1010,4 +1063,9 @@ def _fetch_range(client, bucket, key, start, end, max_attempts=10):
                 return b''
             else:
                 raise
+        except Exception as e:
+            if 'time' in str(e).lower():  # Actual exception type changes often
+                continue
+            else:
+                raise
     raise RuntimeError("Max number of S3 retries exceeded")
